"""
.. module:: ComparativeAmplification.py
    :synopsis: Subroutines to work with the mathematical representation of the Amplitude Amplification
    that uses a double entry oracle (which result depends on 'two' input states and not 'one').
.. moduleauthor::  Ruben Ibarrondo (rubenibarrondo@gmail.com)
"""

import numpy as np
import matplotlib.pyplot as plt
from datetime import datetime
from time import time
import plotTools


# --------------------------------
#
#        Some distributions
#
# --------------------------------

def step(N, minval, maxval, step):
    '''step distribution'''
    bb = np.ones(N)
    if minval>0 and maxval<N:
        bb[minval:maxval] = np.sqrt(step / (maxval - minval))
        bb[:minval] = np.sqrt((1-step) / (N-minval+maxval))
        bb[maxval:] = np.sqrt((1-step) / (N-minval+maxval))
    elif minval == 0 and maxval<N:
        bb[:maxval] = np.sqrt(step / maxval)
        bb[maxval:] = np.sqrt((1-step) / (N+maxval))
    elif maxval == N and minval>0:
        bb[minval:] = np.sqrt(step / (N - minval))
        bb[:minval] = np.sqrt((1-step) / (2*N-minval))
    else:
        bb[:] = 1/np.sqrt(N)
    return bb

def lininc(N):
    '''linearly increasing distribution'''
    return np.array( [ nu for nu in range(N)]) * np.sqrt( 6 / N / (N-1) / (2*N-1))

def lindec(N):
    '''linearly decreasing distribution'''
    return np.array( [ N-1-nu for nu in range(N)]) * np.sqrt( 6 / N / (N-1) / (2*N-1))

def poisson(N, lam):
    '''Poisson distribution'''
    return np.sqrt(np.array( [ lam**nu / np.math.factorial(nu) for nu in range(N)]) * np.exp(-lam))

def discreticed(N, iarr):
    '''Discreticed'''
    bb = np.zeros(N)
    bb[iarr] = 1/np.sqrt(len(iarr))
    return bb

# --------------------------------
#
#       Some test functions
#
# --------------------------------

def test_function(N, aa, funcname, function, z):
    ii = np.arange(N)
    bb = np.zeros(N)
    for nu in range(N):
        for mu in range(N):
            bb[nu] += oracle(mu, z, f=function)
        bb[nu] *= 2 / N
        bb[nu] -= oracle(nu, z, f=function)
        bb[nu] /= np.sqrt(N)

    pp = np.zeros(N)
    for i in ii:
        pp[i] += prob(i, aa, bb, function=function)

    print('| {:12s} | {:12s} | {:12s} | {:12s} | {:12s}|'.format('function', 'z & tz/N', 'bad (f<%20)', 'good (f>%80)',
                                                                 'excel (f>%90)'))
    print('| {:12s} | {:12s} | {:12s} | {:12s} | {:12s} |'.format('-' * 12, '-' * 12, '-' * 12, '-' * 12, '-' * 12))
    fp = [(function(i), pp[i]) for i in ii]
    fp.sort()
    bad = sum(fp[i][1] for i in range(N // 5))
    good = sum(fp[i][1] for i in range(N - N // 5, N))
    excel = sum(fp[i][1] for i in range(N - N // 10, N))
    tz = sum(1 for i in range(N) if function(i) > function(z))
    print('| {:>12s} | {:3} & {:5.4f} | {:12f} | {:12f} | {:12f} |'.format(funcname, z, tz / N, bad, good, excel))

    plt.bar(ii, aa * aa, label='prob_reg1_initial')
    plt.bar(ii, bb * bb, alpha=0.7, label='prob_reg2_threshold')
    plt.bar(ii, pp, alpha=0.7, label='prob_reg1_final')
    plt.legend()
    plt.show()

    plt.title('The function')
    plt.bar(ii, function(ii))
    plt.show()

    plt.plot(function(ii), pp, '.')
    plt.ylabel('probability')
    plt.xlabel('fitness')
    plt.show()


def show_ca_effect():
    N = 64
    z = 15

    function = lambda x: np.sin(2*np.pi*x/N)**2
    a = step(N, minval=0, maxval=N, step=1)
    b = comparative_amplification_oneref(a, z, function=function)

    ab = np.array([[a[i]*b[nu] for nu in range(N)] for i in range(N)])
    ab_new = ab
    ab_list = [ab_new]
    tt = np.arange(10)
    pmax = np.empty(tt.shape)

    for t in tt:
        ab_new = comparative_amplification(ab_new, function=function)
        ab_list.append(ab_new)
        pmax[t] = sum((ab_new * ab_new).transpose())[N-1]
    plt.plot(tt, pmax)
    plt.title('pmax - t')
    plt.show()

    tolerance = 0.01
    expected_steps = -1/pmax * np.log(tolerance)
    plt.plot(tt, expected_steps)
    plt.title('exp. iter. (tolerance={}) - t'.format(tolerance))
    plt.show()

    plotTools.show_finalstate(ab_list[1],
                              row_label='Prob dist in reg1',
                              col_label='Prob dist in reg2',
                              figurelabel='Probability of each state i nu')


# --------------------------------
#
#         Basic operators
#
# --------------------------------

def oracle(i, z, f=lambda x: x):
    if f(i) > f(z):
        return -1
    else:
        return 1

def prob(i, a, b, function=lambda x: x):
    # assumes the given amplitudes are real
    N = len(a)

    p = 0
    for nu in range(N):

        pnu = 0
        for j in range(N):
            pnu += a[j] * oracle(j, nu, function)
        pnu *= 2 / N
        pnu -= a[i] * oracle(i, nu, function)
        pnu *= b[nu]

        p += pnu ** 2
    return p


def prob_compared_amplification(i, z, N, function=lambda x: x):
    # assumes the given amplitudes are real
    a = step(N, minval=0, maxval=N, step=1)
    b = comparative_amplification_oneref(a, z, function=function)

    p = 0
    for nu in range(N):

        pnu = 0
        for j in range(N):
            pnu += a[j] * oracle(j, nu, function)
        pnu *= 2 / N
        pnu -= a[i] * oracle(i, nu, function)
        pnu *= b[nu]

        p += pnu ** 2
    return p


def prob_one_threshold(i, z, N, function=lambda x: x):
    a = 1/np.sqrt(N) * sum(oracle(j, z, function) for j in range(N))
    a *= 2 / N
    a += - 1/np.sqrt(N) * oracle(i, z, function)
    return a*a


def prob_above_threshold(prob_dist, z, function):
    p = 0
    threshold = function(z)
    for i in range(len(prob)):
        if function(i) > threshold:
            p += prob_dist[i]
    return p


def get_above_index(z, N, function):
    abovei = []
    threshold = function(z)
    for i in range(N):
        if function(i) > threshold:
            abovei.append(i)
    return abovei


def get_above_number(z, N, function):
    tz = 0
    threshold = function(z)
    for i in range(N):
        if function(i) > threshold:
            tz += 1
    return tz


def comparative_amplification(ab, function=lambda x: x):
    N, M = ab.shape
    assert N == M, 'ab has to be an square matrix, but {}x{} was obtained'.format(N, M)

    ab_new = np.empty(ab.shape)
    for nu in range(N):
        for i in range(N):
            ab_new[i, nu] = sum(ab[j, nu] * oracle(j, nu, function) for j in range(N))
            ab_new[i, nu] *= 2/N
            ab_new[i, nu] += -ab[i, nu] * oracle(i, nu, function)
    return ab_new

def comparative_amplification_separable(a, b, function=lambda x: x):
    N, M = len(a), len(b)
    assert N == M, 'a and b have to be same length, but {} and {} were obtained'.format(N, M)

    ab_new = np.empty((N, M))
    for nu in range(N):
        for i in range(N):
            ab_new[i, nu] = sum(a[j] * oracle(j, nu, function) for j in range(N))
            ab_new[i, nu] *= 2/N
            ab_new[i, nu] += -a[i] * oracle(i, nu, function)
            ab_new[i, nu] *= b[nu]
    return ab_new

def comparative_amplification_oneref(a, z, function=lambda x: x):
    N = len(a)
    a_new = np.empty(N)
    for i in range(N):
        a_new[i] = sum(a[j] * oracle(j, z, function) for j in range(N))
        a_new[i] *= 2/N
        a_new[i] += -a[i] * oracle(i, z, function)
    return a_new


def compete_dhca_vs_sca(countmax, N=64, function=lambda x: x, zinit=None):
    count = 0
    if not zinit and zinit != 0:
        z = np.random.randint(N)
    else:
        z = zinit
    zdh = z
    zsca = z
    while count < countmax:
        pdh = np.array([prob_one_threshold(i, zdh, N, function) for i in range(N)])
        psca = np.array([prob_compared_amplification(i, zsca, N, function) for i in range(N)])

        yield get_above_number(zdh, N, function)/N, get_above_number(zsca, N, function)/N

        count += 1

        if count < countmax:
            zdh_new = np.random.choice(np.arange(N), p=pdh)
            zsca_new = np.random.choice(np.arange(N), p=psca)
            if function(zdh_new) > function(zdh):
                zdh = zdh_new
            if function(zsca_new) > function(zsca):
                zsca = zsca_new


# --------------------------------
#
#      Comparing DHCA with SCA
#
# --------------------------------

def xz_dhca_vs_sca(countmax, zinit, repetitions, dirpath, N=64):
    filepath = dirpath + '/xz_dhca_vs_sca_c' + str(countmax)\
               + '_z' + str(zinit) + '_r' + str(repetitions) + '_n' + str(N) + '.data'
    with open(filepath, 'a') as f:
        print("-"*32, file=f)
        print("day: ", datetime.now().date(), file=f)
        print("countmax = {}, zinit = {}, repetitions = {}, N = {}".format(countmax, zinit, repetitions, N), file=f)
        t1 = time()
        l = 0
        for rep in range(repetitions):
            ii = []
            jj = []
            for i, j in compete_dhca_vs_sca(countmax=countmax, zinit=zinit, N=N):
                ii.append(i)
                jj.append(j)
            print("xz_dhca_{}".format(l), file=f)
            print(ii, file=f)
            print("xz_sca_{}".format(l), file=f)
            print(jj, file=f)
            print(file=f)
        print('time needed = {}'.format(time() - t1), file=f)
        print("-" * 32, file=f)


def T_dhca_vs_sca(countmax, zinit, repetitions, dirpath, N=64):
    filepath = dirpath + '/T_dhca_vs_sca_c' + str(countmax) +\
               '_z' + str(zinit) + '_r' + str(repetitions) + '_n' + str(N) + '.data'
    with open(filepath, 'a') as f:
        print("-"*32, file=f)
        print("day: ", datetime.now().date(), file=f)
        print("countmax = {}, zinit = {}, repetitions = {}, N={}".format(countmax, zinit, repetitions, N), file=f)
        t1 = time()
        l = 0
        for rep in range(repetitions):
            count_i = 0
            count_j = 0
            for i, j in compete_dhca_vs_sca(countmax=countmax, zinit=zinit, N=N):
    # THIS DOESN'T WORK WITH OTHER FUNCTIONS
                if i != 0:
                    count_i += 1
                if j != 0:
                    count_j += 1

                if i==0 and j==0:
                    break
            print("T_dhca_{}".format(l), file=f)
            print(count_i, file=f)
            print("T_sca_{}".format(l), file=f)
            print(count_j, file=f)
            print(file=f)
        print('time needed = {}'.format(time() - t1), file=f)
        print("-" * 32, file=f)


def parse_T_dhca_vs_sca(repetitions, filepath):
    T_dhca = []
    T_sca = []
    with open(filepath, 'r') as f:
        l = 0
        for line in f:
            if (l > 2) and (l <= (3 + repetitions * 5)):
                if (l-3) % 5 == 1:
                    T_dhca.append(int(line.strip()))
                if (l-3) % 5 == 3:
                    T_sca.append(int(line.strip()))
            else:
                if l == (4 + repetitions * 5):
                    l = -1
            l += 1
    return T_dhca, T_sca


def parse_xz_dhca_vs_sca(repetitions, filepath):
    xz_dhca = []
    xz_sca = []
    with open(filepath, 'r') as f:
        l = 0
        for line in f:
            if (l > 2) and (l <= (3 + repetitions * 5)):
                if (l-3) % 5 == 1:
                    xz_dhca.append(eval(line.strip()))
                if (l-3) % 5 == 3:
                    xz_sca.append(eval(line.strip()))
            else:
                if l == 4 + repetitions * 5:
                    l = -1
            l += 1
    return xz_dhca, xz_sca

if __name__ == '__main__':
    # T_dhca_vs_sca(80, 0, 60, 'data')
    ## preparar un scan de T con N
    av_tdhca = []
    av_tsca = []
    nn = np.array([2, 4, 8, 16, 32, 64, 128])
    for n in nn:
        tdhca, tsca = parse_T_dhca_vs_sca(60, 'data/T_dhca_vs_sca_c80_z0_r60_n{}.data'.format(n))
        tdhca.sort()
        tsca.sort()
        tdhca = np.array(tdhca)
        tsca = np.array(tsca)
        av_tdhca.append(sum(tdhca)/len(tdhca))
        av_tsca.append(sum(tsca) / len(tsca))

        print("N = {}".format(n))
        print('tdhca = {:5.3f} ({:5.3f})'.format(sum(tdhca)/len(tdhca), np.sqrt(sum(tdhca**2)/len(tdhca) - (sum(tdhca)/len(tdhca))**2)))
        print('tsca = {:5.3f} ({:5.3f})'.format(sum(tsca) / len(tsca), np.sqrt(sum(tsca ** 2) / len(tsca) - (sum(tsca) / len(tsca)) ** 2)))
    plt.plot(nn, av_tdhca)
    plt.plot(nn, av_tsca)
    plt.plot(nn, nn/2, '--')
    plt.plot(nn, 3*np.sqrt(nn), '--')
    plt.show()

    #xz_dhca_vs_sca(30, 0, 100, 'data')

    xzdhca, xzsca = parse_xz_dhca_vs_sca(100, 'data/xz_dhca_vs_sca_c30_z0_r100.data')
    a, b = parse_xz_dhca_vs_sca(20, 'data/xz_dhca_vs_sca_c30_z0_r20.data')
    xzdhca += a
    xzsca += b
    xzdhca, xzsca = np.array(xzdhca), np.array(xzsca)
    xz_dhca = sum(xzdhca) / len(xzdhca)
    dev_dhca = np.sqrt(sum(xzdhca ** 2) / len(xzdhca) - xz_dhca ** 2)
    xz_sca = sum(xzsca) / len(xzsca)
    dev_sca = np.sqrt(sum(xzsca ** 2) / len(xzsca) - xz_sca ** 2)
    plt.plot(xz_dhca, 'C0')
    plt.plot(xz_dhca + dev_dhca, '--C0')
    plt.plot(xz_dhca - dev_dhca, '--C0')
    plt.plot(xz_sca, 'C1')
    plt.plot(xz_sca + dev_sca, '--C1')
    plt.plot(xz_sca - dev_sca, '--C1')
    plt.show()